/*
 * Copyright (c) 2005 Michael Deryugin <dem@pikenet.ru>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <postgres.h>
#include <funcapi.h>
#include <fmgr.h>
#include <utils/inet.h>
#include <access/tuptoaster.h>

#include "net_math.h"

#define SETMASK(x,y)	((y) == 0 ? 0 : (x) & (0xffffffff << (32 - (y))))
#define GETBIT(x,y)	(((y) >> (32 - (x))) & 0x1)
#define SETBIT_0(x,y)	((x) & ~(1 << (32 - (y))))
#define SETBIT_1(x,y)	((x) | (1 << (32 - (y))))

#define ip_bits(inetptr)     (((inet_struct *)VARDATA(inetptr))->bits)
#define ip_family(inetptr)   (((inet_struct *)VARDATA(inetptr))->family)
#define ip_type(inetptr)     (((inet_struct *)VARDATA(inetptr))->type)
#define ip_addr(inetptr)     (((inet_struct *)VARDATA(inetptr))->ip_addr)
#define ip_maxbits(inetptr)  (ip_family(inetptr) == PGSQL_AF_INET ? 32 : 128)
#define INET_MEMSIZE(a) (VARHDRSZ + ((char*)ip_addr(a) - (char*)VARDATA(a)) + ip_addrsize(a))

int 	 qget(struct qstring *);
int 	 qgetnext(struct qstring *);
int	 qset(struct qstring *, enum qchars);
int	 qsetprev(struct qstring *, enum qchars);
int	 qput(struct qstring *, enum qchars);
int	 qreplace(struct qstring *, ipaggr *);
int	 qdelblank(struct qstring *);
ipaggr *qcomp(ipaggr *, int, u_long, enum qchars);

int    
qget(struct qstring *qstr)
{
	switch (qstr->qdata[qstr->position]) {
	case 'q':
		return Q;
	case 't':
		return T;
	case 'f':
		return F;
	case ' ':
		return SP;
	default:
		return ESYM;
	}
}

int    
qgetnext(struct qstring *qstr)
{
	int qchr;
	while (qstr->position + 1 < qstr->qlength) {
		qstr->position++;
		qchr = qget(qstr);
		if (qchr == SP)
			continue;
		else
			return qchr;
	}
	return -1;
}

int
qset(struct qstring *qstr, enum qchars qchr)
{
	char *ch;

        if (qstr->position > -1 && qstr->position < qstr->qlength) {
		ch = &qstr->qdata[qstr->position];
		if (*ch != 'q' &&  *ch != 'f' && *ch != 't' &&
		    (qchr == Q || qchr == T || qchr == F))
				qstr->alength++;
                switch (qchr) {
                case Q:
                        *ch = 'q';
                        break;
                case T:
                        *ch = 't';
                        break;
                case F:
                        *ch = 'f';
                        break;
		case SP:
			if (*ch != ' ')
				qstr->alength--;
			*ch = ' ';
			break;
                }
		return 1;
	} else
		return -1;
}

int 
qsetprev(struct qstring *qstr, enum qchars qchr)
{
	while (qstr->position > 0 && qget(qstr) == SP) 
        	qstr->position--;
	if (qget(qstr) != SP)
		return qset(qstr, qchr);
	else
		return -1;
}

int
qputnext(struct qstring *qstr, enum qchars qchr)
{
	if (qstr->position + 1 < qstr->qlength) {
		qstr->position++;
		return qset(qstr, qchr);
	} else
		return -1;
}

int 
qput(struct qstring *qstr, enum qchars qchr)
{
	qstr->position++;
	
	if (qstr->position < qstr->qlength) {
		switch (qchr) {
		case Q:
			qstr->qdata[qstr->position] = 'q';
			break;
		case T:
			qstr->qdata[qstr->position] = 't';
			break;
		case F:
			qstr->qdata[qstr->position] = 'f';
			break;
		case SP:
			break;
		}
	} else 
		return -1;
	return 1;
}

int
qreplace(struct qstring *qstr, ipaggr *aa)
{
	int nlength;
	ipaggr *new;
	void *p;
	
	nlength = qstr->qlength + VARSIZE(aa) - VARHDRSZ - 1;
	new = (ipaggr *) MALLOC(nlength + VARHDRSZ);
	if (new == NULL) {
		log_fatal(NULL);
		return -1;
	}
	VARATT_SIZEP(new) = nlength + VARHDRSZ;
	p = new->data;
	memcpy(p, qstr->qdata, (int4) qstr->position); 
	p += qstr->position;
	memcpy(p, aa->data, VARSIZE(aa) - VARHDRSZ); 
	p += VARSIZE(aa) - VARHDRSZ;
	memcpy(p, qstr->qdata + qstr->position + 1, 
	    qstr->qlength - qstr->position - 1);
	qstr->qstring_data = new;
	qstr->qlength = VARSIZE(new) - VARHDRSZ;
	return 1;
}

int
qdelblank(struct qstring *qstr)
{
	ipaggr *new;
	int4 i, j;
	
	new = (ipaggr *) MALLOC(qstr->alength + VARHDRSZ);
	if (new == NULL) {
		log_fatal(NULL);
		return -1;
	}
	VARATT_SIZEP(new) = qstr->alength + VARHDRSZ;
	for (j = 0, i = 0; j < qstr->alength; j++, i++) {
		while (qstr->qdata[i] != 'q' &&
		    qstr->qdata[i] != 't' &&
		    qstr->qdata[i] != 'f')
			i++;	
		new->data[j] = qstr->qdata[i];
	}
	qstr->qstring_data = new;
	qstr->qlength = VARSIZE(new) - VARHDRSZ;
	return 1;
}


#define C_DB	0x80000000
#define C_S0	0x40000000
#define C_S1	0x20000000
#define C_MASK	0x10000000
#define C_PUSH	0x08000000
#define C_POP	0x04000000
#define C_STATE	0x000000FF

enum states {lq, lt, lf, rq, rt, rf, fin};

/* States of Q-tree:
 * lq - node Q is on left subtree 
 * lt - leaf T is on left subtree
 * lf - leaf F is on left subtree
 * rq - node Q is on right subtree
 * rt - leaf T is on right subtree
 * rf - leaf F is on right subtree
 * 
 *   parent node -> Q - X <- left {node,leaf}
 *                    \
 *                      Y <- right {node,leaf}
 */

static const u_int32_t automat[fin+1][5] =
{/*          Q                  T                F                 state */
 {C_PUSH|C_MASK|C_S0|lq , C_MASK|C_S0|lt , C_MASK|C_S0|lf , fin} /*  lq  */
,{              C_S1|rq ,   C_DB|C_S1|fin,        C_S1|rf , fin} /*  lt  */
,{              C_S1|rq ,        C_S1|rt ,   C_DB|C_S1|fin, fin} /*  lf  */
,{C_PUSH|C_MASK|C_S0|lq , C_MASK|C_S0|lt , C_MASK|C_S0|lf , fin} /*  rq  */
,{        C_POP|C_S1|rq ,  C_POP|C_S1|rt ,  C_POP|C_S1|rf , fin} /*  rt  */
,{        C_POP|C_S1|rq ,  C_POP|C_S1|rt ,  C_POP|C_S1|rf , fin} /*  rf  */
,{                   fin,             fin,             fin, fin} /*  fin */
};

int 
qwalk(struct qstate *it, int action, int s_mask, u_long s_addr)
{
	struct qstring *qstr;
	u_int32_t control = 0;

	retry:
	if (action == AGGREGATE) {
		qstate_init(NULL, it);
	}
	qstr = &(it->qstr);
	while (it->state < fin) {
		if ((it->qc = qgetnext(qstr)) == -1) {
			log_fatal("qwalk: unexpected end of qstring");
			return -1;
		}; 
		control = automat[it->state][it->qc];
		
                if (control & C_MASK) {
			if (it->mask < 32)
				it->mask++;
			else {
				log_fatal(
				    "qwalk: depth of qstring must be <= 32");
				return -1;
			}
		}
		if ((control & C_PUSH) && it->mask > 0 && it->sp < STACK_SIZE)
			it->stack[it->sp++] = it->mask;
		if ((control & C_POP) && it->sp >= 0) {
			it->mask = it->stack[--it->sp];
			it->addr = SETMASK(it->addr, it->mask);
		}
		if ((control & C_S0) && it->mask > 0)
			it->addr = SETBIT_0(it->addr, it->mask);
		if ((control & C_S1) && it->mask > 0)
			it->addr = SETBIT_1(it->addr, it->mask);
		if ((control & C_DB) && (action != AGGREGATE)) {
			log_warn("qwalk: not aggregated qstring");
			return -1;
		}
		it->state = control & C_STATE;
		if ((it->qc != Q) && 
		    (it->addr == SETMASK(0xffffffff, it->mask)))
			it->state = fin;
		if ((it->qc != Q) && (it->mask == 0))
			it->state = fin;
		switch (action) {
		case SEARCH_T:
			if (it->qc == T)
				return 1;
			break;
		case SEARCH_F:
			if (it->qc == F)
				return 1;
			break;
		case SEARCH_ADDR:
			if (it->qc != Q && it->mask == 0)
				return 1;
			if (s_addr == it->addr && s_mask == it->mask)
				return 1;
			if (it->qc != Q && 
			    it->addr == SETMASK(s_addr, it->mask))
				return 1;
			break;
		case AGGREGATE:
			if (control & C_DB) {
				enum qchars cur;
				cur = it->qc;
				qset(&it->qstr, SP);
				qsetprev(&it->qstr, SP);
				qsetprev(&it->qstr, cur);
				goto retry;
			}
			break;
		case SET_BLANK:
			if (it->mask > s_mask)
				qset(&it->qstr, SP);
			else 
				return 1;
			break;
		}
	}
	return 0;
}

void
qstate_init(ipaggr *aa, struct qstate *it)
{
	struct qstring *qstr = &it->qstr;
	
	qstr->position = -1;
	it->addr = 0x00000000;
	it->mask = -1;
	it->state = lq;
	it->sp = 0;
	if (aa != NULL) {
		qstr->qstring_data = aa;
		qstr->qlength = VARSIZE(aa) - VARHDRSZ;
		qstr->alength = qstr->qlength;
	}
}

ipaggr *
qgenerate(int level, int mask, u_long addr, enum qchars type)
{
	struct qstring qstr;
	ipaggr *aa;
	int4 length;
	int i = level;
	int pos = 0;
	enum qchars invtype = T;

	if (type == T)
		invtype = F;
	if (level > mask) {
		log_fatal("qgenerate: level must be <= mask");
		return NULL;
	}
	if (mask > 32) {
		log_fatal("qgenerate: mask must be <= 32");
		return NULL;
	}
	length = 1 + (mask - level) * 2;
	aa = (ipaggr *) MALLOC(length + VARHDRSZ);
	VARATT_SIZEP(aa) = length + VARHDRSZ;
	if (aa == NULL) {
		log_fatal(NULL);
		return NULL;
	}
	
	qstr.qstring_data = aa;
	qstr.qlength = length;
	qstr.alength = length;
	qstr.position = -1;
	
	if (i == mask) {
		qputnext(&qstr, type);
		return aa;
	} else
		qputnext(&qstr, Q);
	pos++;
	i++;
	for (; i < mask; i++) {
		if (GETBIT(i, addr) == 0) {
			qputnext(&qstr, Q);
			pos++;
		} else {
			qputnext(&qstr, invtype);
			qputnext(&qstr, Q);
			pos += 2;
		}	
	}
	if (GETBIT(i, addr) == 0) {
		qputnext(&qstr, type);
		qputnext(&qstr, invtype);
	} else {
		qputnext(&qstr, invtype);
		qputnext(&qstr, type);
	}
	pos += 2;
	for (; pos < length; pos++)
		qputnext(&qstr, invtype);
	return aa;
}

ipaggr *
qcomp(ipaggr *aa, int mask, u_long addr, enum qchars type)
{
	struct qstate qu;
	struct qstate *qs = &qu;
	enum qchars invtype = T;

	if (type == T)
		invtype = F;
	if (aa == NULL) {
		if (type == T)
			return qgenerate(0, mask, addr, T);
		else
			return NULL;
	}
	qstate_init(aa, qs);
	if (qwalk(qs, SEARCH_ADDR, mask, addr) == 1) {
		if (qs->qc == type) 
			return aa;
		if (qs->mask != mask) {
			ipaggr *naa;
			naa = qgenerate(qs->mask, mask, addr, type);
			if (naa == NULL)
				return NULL;
			if (qreplace(&qs->qstr, naa) != 1) {
				FREE(naa);
				return NULL;
			} else {
				FREE(naa);
				return qs->qstr.qstring_data;
			}
			/* NOTREACHED */
		} else if (qs->qc == invtype)
			qset(&qs->qstr, type);
		else if (qs->qc == Q) {
			qset(&qs->qstr, type);
			qwalk(qs, SET_BLANK, qs->mask, 0);
		}
		qwalk(qs, AGGREGATE, 0, 0);
		if (qs->qstr.alength != qs->qstr.qlength)	
			qdelblank(&qs->qstr);
		return qs->qstr.qstring_data;
	}
	return NULL;
} 

ipaggr *
qintersect(ipaggr *aa, ipaggr *bb)
{
	struct qstate qu,qy;
	struct qstate *qs = &qu, *qp = &qy;
	ipaggr *res = NULL, *old = NULL;	

	if (aa == NULL || bb == NULL)
		return NULL;
	qstate_init(aa, qs);
	qstate_init(bb, qp);
	
	if (qwalk(qp, SEARCH_T, 0, 0) <= 0 || qwalk(qs, SEARCH_T, 0, 0) <= 0)
		goto end;
	while (1) {
		while (qs->mask <= qp->mask &&
		    qs->addr == SETMASK(qp->addr, qs->mask) && 
		    qs->qc == T && qp->qc == T) {
			res = qadd(res, qp->mask, qp->addr);
			if (old != res && old != NULL)
				FREE(old);
			old = res;
			if (qwalk(qp, SEARCH_T, 0, 0) <= 0)
				goto end;
		}
		while (qp->mask <= qs->mask &&
		    qp->addr == SETMASK(qs->addr, qp->mask) &&
		    qs->qc == T && qp->qc == T) {
			res = qadd(res, qs->mask, qs->addr);
			if (old != res && old != NULL)
				FREE(old);
			old = res;
			if (qwalk(qs, SEARCH_T, 0, 0) <= 0)
				goto end;
		}
		if (qs->addr > qp->addr) {
			if (qwalk(qp, SEARCH_T, 0, 0) <= 0)
				goto end;
		} else {
			if (qwalk(qs, SEARCH_T, 0, 0) <= 0)
				goto end;
		}
	}
	end:
	if (res == NULL)
		res = qgenerate(0, 0, 0, F);
	return res;
}

ipaggr *
qadd(ipaggr *aa, int mask, u_long addr)
{
	return qcomp(aa, mask, addr, T);
}

ipaggr *
qsub(ipaggr *aa, int mask, u_long addr)
{
	return qcomp(aa, mask, addr, F);
}

int
qcheck(ipaggr *aa)
{
	struct qstate qu;
	struct qstate *qs = &qu;

	qstate_init(aa, qs);
	/* FIX magic number */
	if (qwalk(qs, SEARCH_ADDR, 32, 0xffffffff) > 0 && 
	    qs->qstr.position == (qs->qstr.qlength - 1)) {
		return 1;
	}
	else {
		ereport(ERROR,
		    (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
		    errmsg("ipaggr: full aggregate is smaller than given")));
		return 0;
	}
}

void
log_warn(const char *emsg)
{
     ereport(ERROR,
	    (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
       	     errmsg(emsg)));
}

void
log_fatal(const char *emsg) 
{
	if (emsg == NULL ) {
		ereport(ERROR,
		    (errcode(ERRCODE_OUT_OF_MEMORY),
		    errmsg("ipaggr: out of memory")));
	} else {
	      	ereport(ERROR,
      		    (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
   	    	    errmsg(emsg)));
	}
}

PG_FUNCTION_INFO_V1(ipaggr_in);

Datum
ipaggr_in(PG_FUNCTION_ARGS)
{
	char *cstr = (char *) PG_GETARG_CSTRING(0);
	char *p = cstr;
	int4 len = 0;
	ipaggr *result;

	len = strlen(cstr);
	while (*p) {
		switch (*p) {
		case 'q':
		case 't':
		case 'f':
			break;
		default:	
			ereport(ERROR,
			    (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
			    errmsg("invalid input syntax for ipaggr")));
		}
		p++;
	}
	result = (ipaggr *) palloc(len + VARHDRSZ);
	if (result == NULL)
		ereport(ERROR, 
		    (errcode(ERRCODE_OUT_OF_MEMORY),
		    errmsg("ipaggr: out of memory")));
	memcpy(VARDATA(result), cstr, len);
	VARATT_SIZEP(result) = len + VARHDRSZ;
	if (qcheck(result) == 1)
		PG_RETURN_POINTER(result);
	PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(ipaggr_out);

Datum
ipaggr_out(PG_FUNCTION_ARGS)
{
	ipaggr *aggr;
	char *cstr;
	int4 len;

	aggr = (ipaggr *) PG_GETARG_POINTER(0);
	aggr = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr);
	len = VARSIZE(aggr) - VARHDRSZ;
	cstr = (char *) palloc(len + 1);
	if (cstr == NULL)
		ereport(ERROR,
		    (errcode(ERRCODE_OUT_OF_MEMORY),
		    errmsg("ipaggr: out of memory")));
	memcpy(cstr, VARDATA(aggr), len);
	cstr[len] = '\0';
	PG_RETURN_CSTRING(cstr);
}

static int
ip_addrsize(inet *inetptr)
{
	switch (ip_family(inetptr)) {
	case PGSQL_AF_INET:
		return 4;
	case PGSQL_AF_INET6:
		return 16;
	default:
		return -1;
	}
}

u_long
ip_v4addr(inet *inetptr)
{
	int i = 0;
	u_long a = 0;
	u_char *p;

	for (p = ip_addr(inetptr); i < ip_addrsize(inetptr); i++) {
		a = (a << 8) | p[i];
	}
	return(a);
}

void
toip_v4addr(inet *inetptr, u_long a)
{
	int i;
	u_char *p;

	for (p = ip_addr(inetptr), i = ip_addrsize(inetptr) - 1; i >= 0; i--) {
		p[i] = a;
		a >>= 8;
	}
}

PG_FUNCTION_INFO_V1(check_net);

Datum
check_net(PG_FUNCTION_ARGS)
{
	ipaggr *aggr;
	inet *net;
	u_long s_addr;
	int s_mask;
	struct qstate qu;
	struct qstate *qs = &qu;

	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		PG_RETURN_NULL();
	
	aggr = (ipaggr *) PG_GETARG_POINTER(0);
	aggr = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr);
	qstate_init(aggr, qs);

	net = PG_GETARG_INET_P(1);
	if (ip_family(net) == PGSQL_AF_INET6)
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		    errmsg("add_net: address family AF_INET6 not supported yet")));
	s_mask = ip_bits(net);
	s_addr = ip_v4addr(net);

	if (qwalk(qs, SEARCH_ADDR, s_mask, s_addr) > 0 && qs->qc == T)
		PG_RETURN_BOOL(true);
	else
		PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(add_net);

Datum
add_net(PG_FUNCTION_ARGS)
{
	ipaggr *aggr, *newaggr;
	inet *net;
	u_long s_addr;
	int s_mask;
	
	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		 PG_RETURN_NULL();
	aggr = (ipaggr *) PG_GETARG_POINTER(0);
	aggr = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr);
	net = PG_GETARG_INET_P(1);
	if (ip_family(net) == PGSQL_AF_INET6)
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		    errmsg("add_net: address family AF_INET6 not supported yet")));
	s_mask = ip_bits(net);
	s_addr = ip_v4addr(net);
	newaggr = qadd(aggr, s_mask, s_addr);
	PG_RETURN_POINTER(newaggr);
}

PG_FUNCTION_INFO_V1(sub_net);

Datum
sub_net(PG_FUNCTION_ARGS)
{
	ipaggr *aggr, *newaggr;
	inet *net;
	u_long s_addr;
	int s_mask;
	
	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		 PG_RETURN_NULL();
	aggr = (ipaggr *) PG_GETARG_POINTER(0);
	aggr = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr);
	net = PG_GETARG_INET_P(1);
	if (ip_family(net) == PGSQL_AF_INET6)
		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		    errmsg("add_net: address family AF_INET6 not supported yet")));
	s_mask = ip_bits(net);
	s_addr = ip_v4addr(net);
	newaggr = qsub(aggr, s_mask, s_addr);
	PG_RETURN_POINTER(newaggr);
}

PG_FUNCTION_INFO_V1(add_ipaggr);

Datum
add_ipaggr(PG_FUNCTION_ARGS)
{
	ipaggr *aggr1, *aggr2, *new, *old;
	struct qstate qu;
	struct qstate *qs = &qu;
	
	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		 PG_RETURN_NULL();
	aggr1 = (ipaggr *) PG_GETARG_POINTER(0);
	aggr1 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr1);
	aggr2 = (ipaggr *) PG_GETARG_POINTER(1);
	aggr2 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr2);
	qstate_init(aggr2, qs);
	new = aggr1;
	old = aggr1;
	while (qwalk(qs, SEARCH_T, 0, 0) > 0) {
		new = qadd(new, qs->mask, qs->addr);
		if (old != aggr1 && old != new)
			FREE(old);
		old = new;
	}
	PG_RETURN_POINTER(new);
}

PG_FUNCTION_INFO_V1(sub_ipaggr);

Datum
sub_ipaggr(PG_FUNCTION_ARGS)
{
	ipaggr *aggr1, *aggr2, *new, *old;
	struct qstate qu;
	struct qstate *qs = &qu;
	
	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		 PG_RETURN_NULL();
	aggr1 = (ipaggr *) PG_GETARG_POINTER(0);
	aggr1 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr1);
	aggr2 = (ipaggr *) PG_GETARG_POINTER(1);
	aggr2 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr2);
	qstate_init(aggr2, qs);
	new = aggr1;
	old = aggr1;
	while (qwalk(qs, SEARCH_T, 0, 0) > 0) {
		new = qsub(new, qs->mask, qs->addr);
		if (old != aggr1 && old != new)
			FREE(old);
		old = new;
	}
	PG_RETURN_POINTER(new);
}

PG_FUNCTION_INFO_V1(inter_ipaggr);

Datum
inter_ipaggr(PG_FUNCTION_ARGS)
{
	ipaggr *aggr1, *aggr2, *new;
	
	if (PG_ARGISNULL(0) || PG_ARGISNULL(1))
		PG_RETURN_NULL();
	aggr1 = (ipaggr *) PG_GETARG_POINTER(0);
	aggr1 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr1);
	aggr2 = (ipaggr *) PG_GETARG_POINTER(1);
	aggr2 = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr2);
	new = qintersect(aggr1, aggr2);
	PG_RETURN_POINTER(new);
}

PG_FUNCTION_INFO_V1(show_net);

Datum
show_net(PG_FUNCTION_ARGS)
{
	inet *net;
	ipaggr *aggr;
	struct qstate *qs;

	FuncCallContext  *funcctx;
	
	if (SRF_IS_FIRSTCALL()) {
		MemoryContext oldcontext;
		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		if (PG_ARGISNULL(0))
			SRF_RETURN_DONE(funcctx);
		aggr = (ipaggr *) PG_GETARG_POINTER(0);
		aggr = (ipaggr *) heap_tuple_untoast_attr((varattrib *) aggr);
		qs = (struct qstate *) palloc(sizeof(struct qstate));
		if (qs == NULL)
			ereport(ERROR,
			    (errcode(ERRCODE_OUT_OF_MEMORY),
			    errmsg("show_net: out of memory")));
		MemSet(qs, 0, sizeof(struct qstate));
		qstate_init(aggr, qs);
		funcctx->user_fctx = (void *) qs;
		MemoryContextSwitchTo(oldcontext);
	}
	funcctx = SRF_PERCALL_SETUP();
	qs = (struct qstate *) funcctx->user_fctx;
	aggr = qs->qstr.qstring_data;
	if (qwalk(qs, SEARCH_T, 0, 0) > 0) {
		net = (inet *) palloc(sizeof(inet_struct) + VARHDRSZ);
		if (net == NULL)
			ereport(ERROR,
			    (errcode(ERRCODE_OUT_OF_MEMORY),
			    errmsg("show_net: out of memory")));
		MemSet(net, 0, sizeof(inet_struct) + VARHDRSZ);
		ip_family(net) = PGSQL_AF_INET;
		ip_bits(net) = qs->mask;
		ip_type(net) = 0;
		toip_v4addr(net, qs->addr);
		VARATT_SIZEP(net) = sizeof(inet_struct) + VARHDRSZ;
		funcctx->user_fctx = (void *) qs;
		SRF_RETURN_NEXT(funcctx, InetPGetDatum(net));
	} else {
		SRF_RETURN_DONE(funcctx);
	}
}
